import sys
import os
import pytest
from liquidhandling import SoloSoft


class TestSoloSoft:

    example_pipeline = [
        ["GetTip", "Position1", "TipDisposal", 8, 1, 0, False, "!@#$"],
        ["Loop", 1, "!@#$"],
        [
            "Aspirate",
            "Position1",
            0,
            2,
            100,
            1,
            True,
            False,
            True,
            False,
            "Position1",
            [0, 0, 0],
            0,
            [0, 0, 0],
            "",
            1,
            0,
            0,
            0,
            0,
            0,
            "a",
            0,
            0,
            0,
            0,
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ],
            0,
            0,
            5,
            0,
            0,
            0,
            [0, 0, 0],
            "!@#$",
        ],
        [
            "Dispense",
            "Position1",
            0,
            2,
            100,
            0,
            True,
            False,
            True,
            False,
            "Position1",
            [0, 0, 0],
            0,
            [0, 0, 0],
            "",
            1,
            0,
            0,
            0,
            0,
            0,
            "a",
            0,
            0,
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ],
            0,
            0,
            [0, 0, 0],
            "!@#$",
        ],
        ["EndLoop", "!@#$"],
        ["MoveArm", "TipDisposal", 100, 1, "!@#$"],
    ]

    example_platelist = [
        "TipBox-Corning 200uL",
        "Corning 3383",
        "Corning 3383",
        "Corning 3383",
        "Corning 3383",
        "Corning 3383",
        "Corning 3383",
        "Corning 3383",
    ]

    def test_empty(self):
        soloSoft = SoloSoft.SoloSoft()
        assert soloSoft.filename == None
        assert soloSoft.pipeline == []
        assert soloSoft.plateList == [
            "Empty",
            "Empty",
            "Empty",
            "Empty",
            "Empty",
            "Empty",
            "Empty",
            "Empty",
        ]

    def test_filename(self):
        soloSoft = SoloSoft.SoloSoft("example_filename.hso")
        assert soloSoft.filename == "example_filename.hso"
        soloSoft.setFile("new_filename.hso")
        assert soloSoft.filename == "new_filename.hso"
        with pytest.raises(TypeError):
            soloSoft.setFile(1)
        soloSoft.filename = "once_more.hso"
        assert soloSoft.filename == "once_more.hso"

    def test_pipeline(self):
        soloSoft = SoloSoft.SoloSoft(pipeline=self.example_pipeline)
        assert soloSoft.pipeline == self.example_pipeline
        soloSoft.setPipeline([])
        assert soloSoft.pipeline == []
        soloSoft.setPipeline(self.example_pipeline)
        assert soloSoft.pipeline == self.example_pipeline
        soloSoft.pipeline = []
        assert soloSoft.pipeline == []
        soloSoft.pipeline = self.example_pipeline
        assert soloSoft.pipeline == self.example_pipeline
        with pytest.raises(TypeError):
            soloSoft.setPipeline(3)

    def test_plates(self):
        soloSoft = SoloSoft.SoloSoft(plateList=self.example_platelist)
        assert soloSoft.plateList == self.example_platelist
        soloSoft.setPlates([])
        assert soloSoft.plateList == []
        soloSoft.setPlates(self.example_platelist)
        assert soloSoft.plateList == self.example_platelist
        soloSoft.plateList = []
        assert soloSoft.plateList == []
        soloSoft.plateList = self.example_platelist
        assert soloSoft.plateList == self.example_platelist
        with pytest.raises(TypeError):
            soloSoft.setPlates(3)



from .SoloSoft import SoloSoft
from .Properties import Properties
from .SoftLinx import SoftLinx
from .VolumeManager import VolumeManager
from .Plates import (
    PlateDefinition,
    GenericPlate96Well,
    NinetySixDeepWell,
    NinetySixPlateOneVBottom,
    ZAgilentReservoir_1row,
)



import os
import xml.etree.ElementTree as ET


class SoftLinx:
    def __init__(
        self,
        displayName=None,
        filename=None,
        protocolSteps=None,
        variables=None,
        plates=None,
    ):
        self.displayName = None
        self.filename = None
        self.protocolSteps = []
        self.variables = []
        self.plates = {}
        self.plugin_flags = {
            "PlateCrane": False,
            "Plates": False,
            "Solo": False,
            "RapidPick": False,
            "TorreyPinesRIC20": False,
        }
        self.plugin_reference = {
            "PlateCrane": 0,
            "Plates": 1,
            "Solo": 2,
            # "RapidPick": 3,
            # "TorreyPinesRIC20": 4,
        }
        self.plugin_address = {
            "PlateCrane": 3,
            "Plates": 4,
            "Solo": 5,
            # "RapidPick": 8,
            # "TorreyPinesRIC20": 9,
        }

        # *Set Protocol Name
        try:
            if filename != None:
                self.setDisplayName(displayName)
            else:
                self.setDisplayName("NewProtocol")
        except Exception as error:
            print("Error setting protocol display name.")
            print(error)
        # *Open protocol file for editing
        try:
            if filename != None:
                self.setFile(filename)
        except Exception as error:
            print("Error creating SoftLinx protocol with filename %s: " % filename)
            print(error)
            return
        # *Initialize Protocol Steps
        try:
            if protocolSteps != None:
                self.setProtocolSteps(protocolSteps)
            else:
                self.setProtocolSteps([])
        except Exception as error:
            print("Error setting protocol steps.")
            print(error)
            return
        # *Initialize Variables
        try:
            if variables != None:
                self.setVariables(variables)
            else:
                self.setVariables([])
        except Exception as error:
            print("Error setting variables.")
            print(error)
            return
        # *Initialize Plates
        try:
            if plates != None:
                self.setPlates(plates)
            else:
                self.setPlates({})
        except Exception as error:
            print("Error setting plates.")
            print(error)
            return

    def setDisplayName(self, displayName):
        if not isinstance(displayName, str):
            raise TypeError("Display Name must be a string")
        self.displayName = displayName

    def setFile(self, filename):
        if not isinstance(filename, str):
            raise TypeError("filename must be a string.")
        self.filename = filename

    def setProtocolSteps(self, protocolSteps):
        if not isinstance(protocolSteps, list):
            raise TypeError("Protocol Steps must be a List.")
        self.protocolSteps = protocolSteps

    def setVariables(self, variables):
        if not isinstance(variables, list):
            raise TypeError("Variables must be a List.")
        self.variables = variables

    def setPlates(self, plates):
        if not isinstance(plates, dict):
            raise TypeError(
                "Plates must be a dict with key 'position' and value 'plate name'."
            )
        self.plates = plates

    def activityCapacityCalculator(self):
        if len(self.protocolSteps) > 0:
            return "4"
        else:
            return "0"

    def generatePluginVariables(self, parentXML, system):
        variable = ET.SubElement(
            parentXML,
            "hwab:Variable",
            {
                "x:TypeArguments": "hwab:Interface",
                "Value": "{x:Reference __ReferenceID%d}"
                % self.plugin_reference[system],
                "x:Key": "SoftLinx." + system,
                "Name": "SoftLinx." + system,
                "Prompt": "False",
            },
        )
        default = ET.SubElement(variable, "hwab:Variable.Default")
        if system == "Plates" and len(self.plates):
            interface = ET.SubElement(
                default,
                "hwab:Interface",
                {
                    "x:Name": "__ReferenceID" + str(self.plugin_reference[system]),
                    "AddinType": system,
                },
            )
            interface_address = ET.SubElement(interface, "hwab:Interface.Address")
            setupData = ET.SubElement(interface, "hwab:Interface.SetupData")
            array = ET.SubElement(setupData, "x:Array", {"Type": "x:String"})
            for key, value in self.plates.items():
                element = ET.SubElement(array, "x:String")
                element.text = key + "=" + value
        else:
            interface = ET.SubElement(
                default,
                "hwab:Interface",
                {
                    "SetupData": "{x:Null}",
                    "x:Name": "__ReferenceID" + str(self.plugin_reference[system]),
                    "AddinType": system,
                },
            )
            interface_address = ET.SubElement(interface, "hwab:Interface.Address")
        sladdress = ET.SubElement(
            interface_address,
            "hcc:SLAddress",
            {
                "x:Name": "__ReferenceID%d" % self.plugin_address[system],
                "Name": system,
                "Workcell": "SoftLinx",
            },
        )

    def generatePluginInterface(self, parentXML, system):
        ref = ET.SubElement(parentXML, "x:Reference")
        ref.text = "__ReferenceID%d" % self.plugin_reference[system]
        key = ET.SubElement(ref, "x:Key")
        sub_ref = ET.SubElement(key, "x:Reference")
        sub_ref.text = "__ReferenceID%d" % self.plugin_address[system]
        # if system != "Plates" or len(self.plates) == 0:
        #     if self.plugin_flags[system]:
        #         plugin = ET.SubElement(
        #             parentXML,
        #             "hwab:Interface",
        #             {
        #                 "x:Key": "{x:Reference __ReferenceID%d}" % self.plugin_reference[system],
        #                 "Address": "{x:Reference __ReferenceID%d}" % self.plugin_reference[system],
        #                 "SetupData": "{x:Null}",
        #                 "AddinType": system,
        #             },
        #         )
        #         return
        #     else:
        #         plugin = ET.SubElement(
        #             parentXML,
        #             "hwab:Interface",
        #             {
        #                 "x:Key": "{x:Reference __ReferenceID%d}" % self.plugin_reference[system],
        #                 "SetupData": "{x:Null}",
        #                 "AddinType": system,
        #             },
        #         )
        #         address = ET.SubElement(plugin, "hwab:Interface.Address")
        #         sladdress = ET.SubElement(
        #             address,
        #             "hcc:SLAddress",
        #             {
        #                 "x:Name": "__ReferenceID%d" % self.plugin_reference[system],
        #                 "Name": system,
        #                 "Workcell": "SoftLinx",
        #             },
        #         )
        #         return
        # # *Non-empty Plates-specific implementation
        # plugin = ET.SubElement(
        #     parentXML,
        #     "hwab:Interface",
        #     {
        #         "x:Key": "{x:Reference __ReferenceID%d}" % self.plugin_reference[system],
        #         "AddinType": system,
        #     },
        # )
        # address = ET.SubElement(plugin, "hwab:Interface.Address")
        # sladdress = ET.SubElement(
        #     address,
        #     "hcc:SLAddress",
        #     {"x:Name":  "__ReferenceID%d" % self.plugin_reference[system], "Name": system, "Workcell": "SoftLinx"},
        # )
        # setupData = ET.SubElement(plugin, "hwab:Interface.SetupData")
        # array = ET.SubElement(setupData, "x:Array", {"Type": "x:String"})
        # for key, value in self.plates.items():
        #     element = ET.SubElement(array, "x:String")
        #     element.text = key + "=" + value

    def generateAutoHotKey(self, softlinx_filename=None, ahk_filename=None):
        if ahk_filename == None:
            ahk_filename = os.path.splitext(self.filename)[0] + ".ahk"
        if softlinx_filename == None:
            softlinx_filename = self.filename
        if softlinx_filename == None:
            raise Exception(
                "Cannot generate AutoHotKey script without a softlinx filename."
            )
        with open(ahk_filename, "w") as file:
            file.write("#SingleInstance, Force\n")
            file.write("SendMode Input\n")
            file.write("SetWorkingDir, %A_ScriptDir%\n")
            file.write("\n")
            file.write(
                'Run, "C:\Program Files (x86)\Hudson Robotics\SoftLinx V\SoftLinxVProtocolEditor.exe" %A_ScriptDir%\\'
                + softlinx_filename
                + "\n"
            )
            file.write("WinActivate, SoftLinx V\n")
            file.write("Sleep, 5000\n")
            file.write("MouseClick, Left, 300, 45\n")
            file.write("Sleep, 1000\n")
            file.write('if WinActive("Not Saved") {\n')
            file.write("\tSend, {Tab}{Enter}\n")
            file.write("\tSleep, 5000\n")
            file.write("}\n")
            file.write('if WinActive("Start Now?") {\n')
            file.write("\tSend, {Enter}\n")
            file.write("\tSleep, 5000\n")
            file.write("}\n")

    def generateStepXML(self, scg_list, step):
        self.plugin_flags[step["system"]] = True
        activity_dict = {
            "IconPath": "{x:Null}",
            "CommandLine": step["Command"],
            "Description": step["Description"],
            "DisplayName": step["system"],
            "HasConstraints": "False",
            "SLXId": step["SLXId"],
            "ToolTip": step["ToolTip"],
            "UserComments": "",
            "isActive": step["isActive"],
            "isCanceled": "False",
            "isSetup": "True",
        }
        step_xml = ET.SubElement(scg_list, "InstrumentActivity", activity_dict)
        arguments = ET.SubElement(step_xml, "InstrumentActivity.Arguments")
        instrument_arguments = ET.SubElement(
            arguments,
            "InstrumentActivityArguments",
            {
                "Address": "{x:Reference __ReferenceID%d}"
                % self.plugin_address[step["system"]],
                "ResultVariable": "{x:Null}",
                "AddinType": step["system"],
                "Command": step["Command"],
            },
        )
        # iarg_add = ET.SubElement(
        #     instrument_arguments, "InstrumentActivityArguments.Address"
        # )
        # sladdress = ET.SubElement(
        #     iarg_add,
        #     "hcc:SLAddress",
        #     {
        #         "x:Name": "__ReferenceID" + str(self.plugin_address[step["system"]]),
        #         "Name": step["system"],
        #         "Workcell": "SoftLinx",
        #     },
        # )
        iarg_arg = ET.SubElement(
            instrument_arguments, "InstrumentActivityArguments.Arguments"
        )
        arg_list = ET.SubElement(
            iarg_arg,
            "scg:List",
            {"x:TypeArguments": "x:Object", "Capacity": str(len(step["args"]))},
        )
        for arg in step["args"]:
            arg_xml = ET.SubElement(arg_list, arg[0])
            arg_xml.text = arg[1]
        iarg_hWnd = ET.SubElement(
            instrument_arguments, "InstrumentActivityArguments.hWnd"
        )
        s = ET.SubElement(iarg_hWnd, "s:IntPtr")

        time_constraints = ET.SubElement(step_xml, "InstrumentActivity.TimeConstraints")
        constraint_list = ET.SubElement(
            time_constraints,
            "scg:List",
            {
                "x:TypeArguments": "hwab:TimeConstraint",
                "Capacity": "0",
            },
        )
        workflowViewState = ET.SubElement(step_xml, "sap2010:WorkflowViewState.IdRef")
        workflowViewState.text = "InstrumentActivity_1"

    def plateCraneMoveCrane(
        self,
        position="SoftLinx.PlateCrane.Safe",
        isActive=True,
        index=None,
        inplace=True,
    ):
        step = {
            "type": "MoveCrane",
            "Command": "Move Crane",
            "Description": "MoveCrane to '%s'" % position,
            "SLXId": "83279d2c-43f7-4f36-8515-103a2f9a3de9",
            "ToolTip": "MoveCrane to '%s'" % position,
            "isActive": str(isActive),
            "system": "PlateCrane",
            "args": [["x:String", position]],
        }

        if inplace:
            if index != None:
                self.protocolSteps.insert(index, step)
            else:
                self.protocolSteps.append(step)
        return step

    def plateCraneMovePlate(
        self,
        positionsFrom=["SoftLinx.PlateCrane.Stack1"],
        positionsTo=["SoftLinx.PlateCrane.Stack2"],
        isActive=True,
        index=None,
        inplace=True,
    ):
        step = {
            "type": "MovePlate",
            "Command": "Move Plate",
            "Description": "From:      %s" % ",".join(positionsFrom)
            + "&#xD;&#xA;"
            + "To: %s" % ",".join(positionsTo),
            "SLXId": "9dabc6a5-8b05-434b-b7e8-b56f8d4cdb7d",
            "ToolTip": "From:      %s" % ",".join(positionsFrom)
            + "&#xD;&#xA;"
            + "To: %s" % ",".join(positionsTo),
            "isActive": str(isActive),
            "system": "PlateCrane",
            "args": [
                ["x:String", ",".join(positionsFrom)],
                ["x:String", ",".join(positionsTo)],
                ["x:String", "Safe"],
                ["x:String", "0"],
                ["x:String", " "],
                ["x:String", "False"],
                ["x:String", "False"],
                ["x:String", "False"],
                ["x:String", " "],
            ],
        }

        if inplace:
            if index != None:
                self.protocolSteps.insert(index, step)
            else:
                self.protocolSteps.append(step)
        return step

    def soloSoftRun(
        self, filename="protocol.hso", isActive=True, index=None, inplace=True
    ):
        step = {
            "type": "Run",
            "Command": "Run",
            "Description": "Protocol: " + filename,
            "SLXId": "5eb5e609-1660-4c5b-ade7-f51828196e13",
            "ToolTip": "Protocol: " + filename,
            "isActive": str(isActive),
            "system": "Solo",
            "args": [
                ["x:String", filename],
            ],
        }

        if inplace:
            if index != None:
                self.protocolSteps.insert(index, step)
            else:
                self.protocolSteps.append(step)
        return step

    def saveProtocol(self, filename=None, generate_ahk=True):
        if filename == None:
            if self.filename != None:
                filename = self.filename
            else:
                raise BaseException("Need to specify a filename to save a protocol.")

        if len(self.plates) > 0:
            self.plugin_flags["Plates"] = True

        # *Start Constructing XML
        protocol_dict = {
            "mc:Ignorable": "sap sap2010 sads",
            "PreProtocolWizard": "{x:Null}",
            "ActivityLabel": "",
            "DisplayName": self.displayName,
            "HasConstraints": "False",
            "sap2010:WorkflowViewState.IdRef": "Protocol_1",
            "SLXId": "df50c4dd-572a-41df-a285-0caf0907a053",
            "ToolTip": "",
            "UserComments": "",
            "isActive": "True",
            "isSetup": "True",
            "xmlns": "clr-namespace:Hudson.Workflow.Activities;assembly=Hudson.Workflow.Activities",
            "xmlns:hcc": "clr-namespace:Hudson.Common.Communications;assembly=Hudson.Common",
            "xmlns:hwab": "clr-namespace:Hudson.Workflow.Activities.Base;assembly=SoftLinxBaseActivities",
            "xmlns:mc": "http://schemas.openxmlformats.org/markup-compatibility/2006",
            "xmlns:p": "http://schemas.microsoft.com/netfx/2009/xaml/activities",
            "xmlns:s": "clr-namespace:System;assembly=mscorlib",
            "xmlns:sads": "http://schemas.microsoft.com/netfx/2010/xaml/activities/debugger",
            "xmlns:sap": "http://schemas.microsoft.com/netfx/2009/xaml/activities/presentation",
            "xmlns:sap2010": "http://schemas.microsoft.com/netfx/2010/xaml/activities/presentation",
            "xmlns:scg": "clr-namespace:System.Collections.Generic;assembly=mscorlib",
            "xmlns:x": "http://schemas.microsoft.com/winfx/2006/xaml",
        }
        protocol = ET.Element(
            "Protocol",
            protocol_dict,
        )

        # *Activities
        activities = ET.SubElement(protocol, "Protocol.Activities")
        scg_list = ET.SubElement(
            activities,
            "scg:List",
            {
                "x:TypeArguments": "p:Activity",
                "Capacity": "4",
            },
        )
        # *Add each step in the protocol
        for step in self.protocolSteps:
            self.generateStepXML(scg_list, step)
        activities2 = ET.SubElement(protocol, "Protocol.Activities2")
        scg_list = ET.SubElement(
            activities2,
            "scg:List",
            {"x:TypeArguments": "p:Activity", "Capacity": "0"},
        )
        initialValues = ET.SubElement(protocol, "Protocol.InitialValues")
        scg_dict = ET.SubElement(
            initialValues,
            "scg:Dictionary",
            {"x:TypeArguments": "x:String, hwab:Variable"},
        )

        # *Add each plugin
        interfaces = ET.SubElement(protocol, "Protocol.Interfaces")
        scg_dict = ET.SubElement(
            interfaces,
            "scg:Dictionary",
            {"x:TypeArguments": "hcc:SLAddress, hwab:Interface"},
        )
        self.generatePluginInterface(scg_dict, "PlateCrane")
        self.generatePluginInterface(scg_dict, "Plates")
        self.generatePluginInterface(scg_dict, "Solo")
        # self.generatePluginInterface(interfaces, "PlateCrane")
        # self.generatePluginInterface(interfaces, "Plates")
        # self.generatePluginInterface(interfaces, "RapidPick")
        # self.generatePluginInterface(interfaces, "Solo")
        # self.generatePluginInterface(interfaces, "TorreyPinesRIC20")

        # *TimeConstraints
        timeConstraints = ET.SubElement(protocol, "Protocol.TimeConstraints")
        scg_list = ET.SubElement(
            timeConstraints,
            "scg:List",
            {"x:TypeArguments": "hwab:TimeConstraint", "Capacity": "0"},
        )

        # *Variables
        variables = ET.SubElement(protocol, "Protocol.Variables")
        variableList = ET.SubElement(
            variables, "hwab:VariableList", {"SLXHost": "{x:Null}"}
        )
        self.generatePluginVariables(variableList, "PlateCrane")
        self.generatePluginVariables(variableList, "Plates")
        # self.generatePluginVariables(variableList, "RapidPick")
        self.generatePluginVariables(variableList, "Solo")
        # self.generatePluginVariables(variableList, "TorreyPinesRIC20")

        # *Add each variable
        # for variable in self.variables:
        #     variable_xml = ET.SubElement(variableList, variable.xml())

        workflowViewState = ET.SubElement(
            protocol, "sap2010:WorkflowViewState.ViewStateManager"
        )
        viewStateManager = ET.SubElement(workflowViewState, "sap2010:ViewStateManager")
        viewStateData = ET.SubElement(
            viewStateManager,
            "sap2010:ViewStateData",
            {
                "Id": "Protocol_1",
                "sap:VirtualizedContainerService.HintSize": "205,119",
            },
        )
        # !debugSymbol is probably some sort of hashed nonsense, which could be an issue
        # !Update: seems to be constructive rather than hashed. But it also doesn't seem to impact operation atm
        debugSymbol = ET.SubElement(protocol, "sads:DebugSymbol.Symbol")
        debugSymbol.text = (
            "dypDOlxVc2Vyc1xyeWFuZFxEZXZcTmV3UHJvdG9jb2xfcGx1Z2luLnNsdnABAQFqDAEB"
        )
        tree = ET.ElementTree(self.indent(protocol))
        xmlstring = ET.tostring(
            tree.getroot(), method="xml", encoding="unicode"
        ).replace("&amp;", "&")
        with open(filename, "w") as file:
            file.write(xmlstring)
        # *Generate AutoHotKey script
        self.generateAutoHotKey(
            os.path.basename(filename), os.path.splitext(filename)[0] + ".ahk"
        )

    # *Pretty-print XML
    def indent(self, element, level=0):
        i = "\n" + level * "  "
        j = "\n" + (level - 1) * "  "
        if len(element):
            if not element.text or not element.text.strip():
                element.text = i + "  "
            if not element.tail or not element.tail.strip():
                element.tail = i
            for subelem in element:
                self.indent(subelem, level + 1)
            if not element.tail or not element.tail.strip():
                element.tail = j
        else:
            if level and (not element.tail or not element.tail.strip()):
                element.tail = j
        return element



import json
import liquidhandling.Properties

STEP_DELIMITER = "!@#$"


class SoloSoft:
    def __init__(self, filename=None, plateList=None, pipeline=None):
        self.filename = None
        self.plateList = []
        self.pipeline = []

        # *Open protocol file for editing
        try:
            if filename != None:
                self.setFile(filename)
        except Exception as error:
            print("Error creating SoloSoft protocol with filename %s" % filename)
            print(error)
            return
        # *Set plate list
        try:
            if plateList != None:
                self.setPlates(plateList)
            else:
                self.setPlates(
                    [
                        "Empty",
                        "Empty",
                        "Empty",
                        "Empty",
                        "Empty",
                        "Empty",
                        "Empty",
                        "Empty",
                    ]
                )
        except Exception as error:
            print("Error setting Plate List")
            print(error)
            return
        # *Set pipeline, if we're expanding on an existing pipeline
        try:
            if pipeline != None:
                self.setPipeline(pipeline)
            else:
                self.initializePipeline()
        except Exception as error:
            print("Error setting pipeline")
            print(error)

    def setFile(self, filename):
        if not isinstance(filename, str):
            raise TypeError("filename must be a string.")
        else:
            self.filename = filename

    def setPlates(self, plateList):
        if not isinstance(plateList, list):
            raise TypeError("plateList must be a list of strings.")
        else:
            self.plateList = plateList

    def setPipeline(self, pipeline):
        if not isinstance(pipeline, list):
            raise TypeError("pipeline should be a list")
        else:
            self.pipeline = pipeline

    def initializePipeline(self):
        self.setPipeline([])

    def removeStep(self, position=-1):
        try:
            self.pipeline.remove(position)
        except:
            print("Error removing step at position %i in pipeline" % position)

    def savePipeline(self, filename=None, CRLF=True):
        if filename == None:
            if self.filename != None:
                filename = self.filename
            else:
                raise BaseException("Need to specify a file to save pipeline")

        if CRLF:
            newline = "\r\n"
        else:
            newline = ""

        with open(filename, "w", newline=newline) as file:
            for plate in self.plateList:
                file.write(str(plate) + "\n")
            for step in self.pipeline:
                for item in step:
                    if isinstance(item, list):
                        if len(item) > 0 and isinstance(item[0], list):
                            for line in item:
                                for number in line[:-1]:
                                    file.write(str(number))
                                    file.write(",")
                                file.write(str(line[-1]))
                                file.write("\n")
                        else:
                            for number in item:
                                file.write(str(number) + "\n")
                    else:
                        file.write(str(item) + "\n")

    def pipelineToJSON(self, json_file=None, pipeline=None, plateList=None):
        if pipeline != None:
            if not isinstance(pipeline, list):
                raise TypeError("pipeline should be a list")
        else:
            pipeline = self.pipeline
        if plateList != None:
            if not isinstance(plateList, list):
                raise TypeError("platelist should be a list")
        else:
            plateList = self.plateList

        json_data = {}
        json_data["metadata"] = {"spec_version": Properties.SPEC_VERSION}
        json_data["pipeline_type"] = "SoloSoft"
        json_data["platelist"] = plateList
        steps = []
        for step in pipeline:
            # step = pipeline[28]
            # if True:
            step_extraction_function = self.jsonify[step[0]]
            step_data = {}
            step_data["step_definition"] = step_extraction_function(self, step)
            steps.append(step_data)
        json_data["steps"] = steps
        return json_data

    def jsonToPipeline(self, json_data, inplace=True):
        if isinstance(json_data, str):
            json_local = json.loads(json_data)
        elif isinstance(json_data, dict):
            json_local = json_data
        else:
            print("json needs to be either a dict or string.")
            return

        steps = []
        for step in json_local["steps"]:
            params = {}
            step_function = None
            for key in step["step_definition"]:
                if key == "step_type":
                    step_function = self.pipelinify[step["step_definition"][key]]
                else:
                    params[key] = step["step_definition"][key]
            steps.append(step_function(self=self, inplace=False, **params))
            params = {}
        if inplace:
            self.setPipeline = steps
        return steps

    # * SOLOSoft Pipeline Functions

    def getTip(
        self,
        position="Position1",
        disposal="TipDisposal",
        num_tips=8,
        auto_tip_selection=True,
        count_tips_from_last_channel=False,
        index=None,
        inplace=True,
    ):
        properties_list = ["GetTip", position, disposal, num_tips]
        if auto_tip_selection:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend([0, count_tips_from_last_channel, STEP_DELIMITER])
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyGetTip(self, step):
        json_data = {}
        json_data["step_type"] = "GetTip"
        json_data["position"] = step[1]
        json_data["disposal"] = step[2]
        json_data["num_tips"] = step[3]
        json_data["auto_tip_selection"] = step[4]
        json_data["count_tips_from_last_channel"] = step[6]
        return json_data

    def shuckTip(self, disposal="TipDisposal", index=None, inplace=True):
        properties_list = ["ShuckTip", disposal, STEP_DELIMITER]
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyShuckTip(self, step):
        json_data = {}
        json_data["step_type"] = "ShuckTip"
        json_data["disposal"] = step[1]
        return json_data

    def loop(self, iterations=-1, index=None, inplace=True):
        properties_list = ["Loop", iterations, STEP_DELIMITER]
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyLoop(self, step):
        json_data = {}
        json_data["step_type"] = "Loop"
        json_data["iterations"] = step[1]
        return json_data

    def endLoop(self, index=None, inplace=True):
        properties_list = ["EndLoop", STEP_DELIMITER]
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyEndLoop(self, step):
        json_data = {}
        json_data["step_type"] = "EndLoop"
        return json_data

    def aspirate(
        self,
        position="Position1",
        aspirate_volume_to_named_point=False,
        aspirate_volume_single=0,
        syringe_speed=100,
        start_by_emptying_syringe=True,
        increment_column_order=False,
        aspirate_point="Position1",
        aspirate_shift=[0, 0, 0],
        do_tip_touch=False,
        tip_touch_shift=[0, 0, 0],
        file_data_path="",
        multiple_wells=1,
        backlash=0,
        pre_aspirate=0,
        mix_at_start=False,
        mix_cycles=0,
        mix_volume=0,
        dispense_height=0,
        delay_after_dispense=0,
        aspirate_volumes=None,
        dwell_after_aspirate=0,
        find_bottom_of_vessel=False,
        reverse_order=False,
        post_aspirate=0,
        move_while_pipetting=False,
        move_distance=[0, 0, 0],
        index=None,
        inplace=True,
    ):
        properties_list = [
            "Aspirate",
            position,
            aspirate_volume_single,
            2,
            syringe_speed,
        ]
        if start_by_emptying_syringe:
            properties_list.append(1)
        else:
            properties_list.append(0)
        if aspirate_volume_to_named_point:
            properties_list.extend([False, True])
        else:
            properties_list.extend([True, False])
        if increment_column_order:
            properties_list.extend([False, True])
        else:
            properties_list.extend([True, False])
        properties_list.extend([aspirate_point, aspirate_shift])
        if do_tip_touch:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend(
            [tip_touch_shift, file_data_path, multiple_wells, backlash, pre_aspirate]
        )
        if mix_at_start:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend(
            [mix_cycles, mix_volume, "a", 0, 0, dispense_height, delay_after_dispense]
        )
        if aspirate_volumes != None:
            properties_list.append(aspirate_volumes)
        else:
            properties_list.append(
                [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                ]
            )
        properties_list.append(dwell_after_aspirate)
        if find_bottom_of_vessel:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.append(5)  # ? Myterious 1 or 2 digit integer
        if reverse_order:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.append(post_aspirate)
        if move_while_pipetting:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend([move_distance, STEP_DELIMITER])
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyAspirate(self, step):
        json_data = {}
        json_data["step_type"] = "Aspirate"
        json_data["position"] = step[1]
        json_data["aspirate_volume_single"] = step[2]
        json_data["syringe_speed"] = step[4]
        json_data["start_by_emptying_syringe"] = step[5]
        json_data["aspirate_volume_to_named_point"] = step[7]
        json_data["increment_column_order"] = step[9]
        json_data["aspirate_point"] = step[10]
        json_data["aspirate_shift"] = step[11]
        json_data["do_tip_touch"] = step[12]
        json_data["tip_touch_shift"] = step[13]
        json_data["file_data_path"] = step[14]
        json_data["multiple_wells"] = step[15]
        json_data["backlash"] = step[16]
        json_data["pre_aspirate"] = step[17]
        json_data["mix_at_start"] = step[18]
        json_data["mix_cycles"] = step[19]
        json_data["mix_volume"] = step[20]
        json_data["dispense_height"] = step[24]
        json_data["delay_after_dispense"] = step[25]
        json_data["aspirate_volumes"] = step[26]
        json_data["dwell_after_aspirate"] = step[27]
        json_data["find_bottom_of_vessel"] = step[28]
        json_data["reverse_order"] = step[30]
        json_data["post_aspirate"] = step[31]
        json_data["move_while_pipetting"] = step[32]
        json_data["move_distance"] = step[33]
        return json_data

    def dispense(
        self,
        position="Position1",
        dispense_volume_single=0,
        syringe_speed=100,
        backlash=0,
        dispense_volume_to_named_point=False,
        increment_column_order=False,
        dispense_point="Position1",
        dispense_shift=[0, 0, 0],
        do_tip_touch=False,
        tip_touch_shift=[0, 0, 0],
        file_data_path="",
        multiple_wells=1,
        dwell_after_dispense=0,
        blowoff=0,
        mix_at_finish=False,
        mix_cycles=0,
        mix_volume=0,
        aspirate_height=0,
        delay_after_aspirate=0,
        dispense_volumes=None,
        reverse_order=False,
        move_while_pipetting=False,
        move_distance=[0, 0, 0],
        index=None,
        inplace=True,
    ):
        properties_list = [
            "Dispense",
            position,
            dispense_volume_single,
            2,
            syringe_speed,
            backlash,
        ]
        if dispense_volume_to_named_point:
            properties_list.extend([False, True])
        else:
            properties_list.extend([True, False])
        if increment_column_order:
            properties_list.extend([False, True])
        else:
            properties_list.extend([True, False])
        properties_list.extend([dispense_point, dispense_shift])
        if do_tip_touch:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend(
            [
                tip_touch_shift,
                file_data_path,
                multiple_wells,
                dwell_after_dispense,
                blowoff,
            ]
        )
        if mix_at_finish:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend(
            [mix_cycles, mix_volume, "a", aspirate_height, delay_after_aspirate]
        )
        if dispense_volumes != None:
            properties_list.append(dispense_volumes)
        else:
            properties_list.append(
                [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                ]
            )
        if reverse_order:
            properties_list.append(1)
        else:
            properties_list.append(0)
        if move_while_pipetting:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend([move_distance, STEP_DELIMITER])
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyDispense(self, step):
        json_data = {}
        json_data["step_type"] = "Dispense"
        json_data["position"] = step[1]
        json_data["dispense_volume_single"] = step[2]
        json_data["syringe_speed"] = step[4]
        json_data["backlash"] = step[5]
        json_data["dispense_volume_to_named_point"] = step[7]
        json_data["increment_column_order"] = step[9]
        json_data["dispense_point"] = step[10]
        json_data["dispense_shift"] = step[11]
        json_data["do_tip_touch"] = step[12]
        json_data["tip_touch_shift"] = step[13]
        json_data["file_data_path"] = step[14]
        json_data["multiple_wells"] = step[15]
        json_data["dwell_after_dispense"] = step[16]
        json_data["blowoff"] = step[17]
        json_data["mix_at_finish"] = step[18]
        json_data["mix_cycles"] = step[19]
        json_data["mix_volume"] = step[20]
        json_data["dispense_height"] = step[22]
        json_data["delay_after_aspirate"] = step[23]
        json_data["dispense_volumes"] = step[24]
        json_data["reverse_order"] = step[25]
        json_data["move_while_pipetting"] = step[26]
        json_data["move_distance"] = step[27]
        return json_data

    def prime(
        self,
        position="Position1",
        syringe_speed=100,
        fill_syringe=False,
        empty_syringe=True,
        aspirate_volume=False,
        dispense_volume=False,
        volume=0,
        index=None,
        inplace=True,
    ):
        properties_list = [
            "Prime",
            syringe_speed,
            True,  # ? Unclear what this is
            False,  # ? Unclear what this is
            False,  # ? Unclear what this is
            0,  # ? Unclear what this is
            "a",  # ? Unclear what this is
            2,  # ? Unclear what this is
            True,  # ? Unclear what this is
            1,  # ? Unclear what this is
            "*",  # ? Unclear what this is
            volume,
            fill_syringe,
            empty_syringe,
            aspirate_volume,
            dispense_volume,
            "*",
            "*",
            STEP_DELIMITER,
        ]
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyPrime(self, step):
        json_data = {}
        json_data["step_type"] = "Prime"
        json_data["syringe_speed"] = step[1]
        json_data["volume"] = step[11]
        json_data["fill_syringe"] = step[12]
        json_data["empty_syringe"] = step[13]
        json_data["aspirate_volume"] = step[14]
        json_data["dispense_volume"] = step[15]
        return json_data

    def pause(
        self,
        pause_message="",
        allow_end_run=False,
        auto_continue_after=False,
        wait_seconds=0,
        index=None,
        inplace=True,
    ):
        properties_list = ["Pause", pause_message]
        if allow_end_run or auto_continue_after:
            properties_list.append(1)
        else:
            properties_list.append(0)
        if auto_continue_after:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.extend([wait_seconds, STEP_DELIMITER])
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyPause(self, step):
        json_data = {}
        json_data["step_type"] = "Pause"
        json_data["pause_message"] = step[1]
        json_data["allow_end_run"] = step[2]
        json_data["auto_continue_after"] = step[3]
        json_data["wait_seconds"] = step[4]
        return json_data

    def getBottom(
        self,
        position="Position1",
        increment_row_order=True,
        increment_column_order=False,
        output_file_path="",
        wells_per_pass=1,  # * -1 for all
        search_start_distance=0,
        well_list=None,
        index=None,
        inplace=True,
    ):
        properties_list = [
            "GetBottom",
            position,
            increment_row_order,
            increment_column_order,
            output_file_path,
            wells_per_pass,
            search_start_distance,
            5,  # ? Unclear what this is
            5,  # ? Unclear what this is
            "*",  # ? Unclear what this is
            "*",  # ? Unclear what this is
        ]
        if well_list != None:
            properties_list.append(well_list)
        else:
            properties_list.append(
                [
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                    [
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                        False,
                    ],
                ]
            )
        properties_list.append(STEP_DELIMITER)
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyGetBottom(self, step):
        json_data = {}
        json_data["step_type"] = "GetBottom"
        json_data["position"] = step[1]
        json_data["increment_row_order"] = step[2]
        json_data["increment_column_order"] = step[3]
        json_data["output_file_path"] = step[4]
        json_data["wells_per_pass"] = step[5]
        json_data["search_start_distance"] = step[6]
        json_data["well_list"] = step[11]
        return json_data

    def setSpeed(self, xyz_speed=100, index=None, inplace=True):
        properties_list = ["SetSpeed", xyz_speed, STEP_DELIMITER]
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifySetSpeed(self, step):
        json_data = {}
        json_data["step_type"] = "SetSpeed"
        json_data["xyz_speed"] = step[1]
        return json_data

    def moveArm(
        self,
        destination="TipDisposal",
        xyz_speed=100,
        move_z_at_start=True,
        index=None,
        inplace=True,
    ):
        properties_list = ["MoveArm", destination, xyz_speed]
        if move_z_at_start:
            properties_list.append(1)
        else:
            properties_list.append(0)
        properties_list.append(STEP_DELIMITER)
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyMoveArm(self, step):
        json_data = {}
        json_data["step_type"] = "MoveArm"
        json_data["destination"] = step[1]
        json_data["xyz_speed"] = step[2]
        json_data["move_z_at_start"] = step[3]
        return json_data

    def movePlate(
        self,
        source_position="Position1",
        target_position="Position2",
        grip_offset=0,
        index=None,
        inplace=True,
    ):
        properties_list = [
            "MovePlate",
            source_position,
            target_position,
            grip_offset,
            STEP_DELIMITER,
        ]
        if inplace:
            if index != None:
                self.pipeline.insert(index, properties_list)
            else:
                self.pipeline.append(properties_list)
        return properties_list

    def jsonifyMovePlate(self, step):
        json_data = {}
        json_data["step_type"] = "MovePlate"
        json_data["source_position"] = step[1]
        json_data["target_position"] = step[2]
        json_data["grip_offset"] = step[3]
        return json_data

    jsonify = {
        "GetTip": jsonifyGetTip,
        "ShuckTip": jsonifyShuckTip,
        "Loop": jsonifyLoop,
        "EndLoop": jsonifyEndLoop,
        "Aspirate": jsonifyAspirate,
        "Dispense": jsonifyDispense,
        "GetBottom": jsonifyGetBottom,
        "Prime": jsonifyPrime,
        "Pause": jsonifyPause,
        "MoveArm": jsonifyMoveArm,
        "MovePlate": jsonifyMovePlate,
        "SetSpeed": jsonifySetSpeed,
    }

    pipelinify = {
        "GetTip": getTip,
        "ShuckTip": shuckTip,
        "Loop": loop,
        "EndLoop": endLoop,
        "Aspirate": aspirate,
        "Dispense": dispense,
        "GetBottom": getBottom,
        "Prime": prime,
        "Pause": pause,
        "MoveArm": moveArm,
        "MovePlate": movePlate,
        "SetSpeed": setSpeed,
    }



import sys
import os
import pytest
from liquidhandling import ExperimentManager
from liquidhandling import ExperimentParameter


class TestExperimentParameter:
    def test_temp(self):
        assert True



import string


class PlateDefinition:
    def __init__(
        self,
        name="Custom Defined Plate",
        plate=None,
        plate_height=0,
        well_depth=0,
        rows=8,
        columns=12,
        x_offset=0,
        y_offset=0,
        row_spacing=9,
        column_spacing=9,
        comments="",
    ):
        if isinstance(name, str):
            self.name = name
        else:
            raise ValueError("Plate name must be a string.")
        if isinstance(plate_height, (int, float)) and plate_height >= 0:
            self.plate_height = plate_height
        else:
            raise ValueError("Plate Height must be a positive numeric value.")
        if isinstance(well_depth, (int, float)) and well_depth >= 0:
            self.well_depth = well_depth
        else:
            raise ValueError("Well depth must be a positive numeric value.")
        if isinstance(rows, int) and rows > 0:
            self.rows = rows
        else:
            raise ValueError("Rows must be a positive integer.")
        if isinstance(columns, int) and columns > 0:
            self.columns = columns
        else:
            raise ValueError("Columns must be a positive integer.")
        if isinstance(x_offset, (int, float)):
            self.x_offset = x_offset
        else:
            raise ValueError("X_offset must be a numeric value")
        if isinstance(y_offset, (int, float)):
            self.y_offset = y_offset
        else:
            raise ValueError("Y_offset must be a numeric value")
        if isinstance(row_spacing, (int, float)) and row_spacing > 0:
            self.row_spacing = row_spacing
        else:
            raise ValueError("Row Spacing must be a positive numeric value")
        if isinstance(column_spacing, (int, float)) and column_spacing > 0:
            self.column_spacing = column_spacing
        else:
            raise ValueError("Column Spacing must be a positive numeric value")
        if plate == None:
            self.plate = [[0 for j in range(self.columns)] for i in range(self.rows)]
        else:
            self.plate == plate
        self.comments = comments

    def setRow(self, row="A", value=0):
        alpha_vals = list(string.ascii_uppercase)
        if str.upper(row) in alpha_vals[0 : self.rows]:
            index = alpha_vals.index(str.upper(row))
            if float(value) >= 0:
                for i in range(self.columns):
                    self.plate[index][i] = value
            else:
                print("value must be a non-negative number")
                return
        elif row in range(1, self.rows + 1):
            if float(value) >= 0:
                for i in range(12):
                    self.plate[row][i] = value
            else:
                print("value must be a non-negative number")
                return
        else:
            print("row must be a character in the range A-H, or a number 1-8")
            return
        return self.plate

    def setColumn(self, column=1, value=0):
        if column in range(1, self.columns + 1):
            if float(value) >= 0:
                for row in range(self.rows):
                    self.plate[row][column - 1] = value
            else:
                print("value must be a non-negative number")
                return
        else:
            print("column must be a number in the range 1-12 inclusive")
            return
        return self.plate

    def setAll(self, value=0):
        if float(value) >= 0:
            for row in range(self.rows):
                for column in range(self.columns):
                    self.plate[row][column] = value
        else:
            print("value must be a non-negative number")
            return
        return self.plate


class GenericPlate96Well:
    def __new__(cls, plate=None):
        return PlateDefinition(
            "Generic 96-well Plate",
            plate,
            14.5,
            10.7,
            8,
            12,
            0,
            0,
            9,
            9,
            "Based on 96 PlateOne V-Bottom",
        )


class NinetySixPlateOneVBottom:
    def __new__(cls, plate=None):
        return PlateDefinition(
            "96 Plate One V Bottom",
            plate,
            14.5,
            10.7,
            8,
            12,
            0,
            0,
            9,
            9,
            "Based on 96 PlateOne V-Bottom",
        )


class ZAgilentReservoir_1row:
    def __new__(cls, plate=None):
        return PlateDefinition(
            "Z Agilent Reservoir - 1 row", plate, 44.5, 42.3, 8, 12, 0, 0, 9, 9, ""
        )


class NinetySixDeepWell:
    def __new__(cls, plate=None):
        return PlateDefinition("96 Deep Well", plate, 42.0, 38.0, 8, 12, 0, 0, 9, 9, "")
